package address_utils

import (
	"crypto/md5"
	"fmt"
	"gitlab.com/uafrica/go-utils/string_utils"
	"regexp"
	"strings"
)

// MD5HashOfAddress m(E,L,L) - calculates and returns the MD5 hash of the entered address, lat and lng concatenated together. If lat and lng is blank, it is only the hash of the entered address
func MD5HashOfAddress(enteredAddress string, lat *float64, lng *float64) string {
	valueToHash := enteredAddress
	if lat != nil && lng != nil {
		valueToHash += fmt.Sprintf(",%v,%v", *lat, *lng)
	}

	return fmt.Sprintf("%X", md5.Sum([]byte(valueToHash)))
}

// MD5HashOfLowerCaseEnteredAddress m(e) - calculates and returns the MD5 hash of a string after preparing the string properly.
func MD5HashOfLowerCaseEnteredAddress(enteredAddress string) string {
	valueToHash := cleanLowerCaseAddress(enteredAddress)
	return fmt.Sprintf("%X", md5.Sum([]byte(valueToHash)))
}

// cleanLowerCaseAddress makes the entered address lowercase, removes unwanted chars, strip street terms and removes all whitespace
func cleanLowerCaseAddress(enteredAddress string) string {
	// Lowercase.
	enteredAddress = strings.ToLower(enteredAddress)

	// Remove unwanted characters.
	enteredAddress = stripUnwantedCharacters(enteredAddress)

	// Remove ave, str, etc.
	enteredAddress = stripStreetTerms(enteredAddress)

	// Remove all whitespace.
	enteredAddress = string_utils.RemoveAllWhiteSpaces(enteredAddress)

	return enteredAddress
}

func stripStreetTerms(s string) string {
	terms := []string{
		"ave",
		"avenue",
		"str",
		"street",
		"st",
		"straat",
		"lane",
		"ln",
		"laan",
		"road",
		"rd",
		"boulevard",
		"blvd",
		"drive",
		"drv",
		"way",
		"weg",
	}
	re := regexp.MustCompile(`\s+(` + strings.Join(terms, "|") + `)[\s]*`)
	s = re.ReplaceAllString(s, "")

	return s
}

func stripUnwantedCharacters(s string) string {
	re := regexp.MustCompile(`[^\s\w,-]`)
	s = re.ReplaceAllString(s, "")

	return s
}

func CleanZone(oldCountry, oldZone *string) (newCountry, newZone *string) {
	// Google zones
	/*
		"long_name": "KwaZulu-Natal",
		"short_name": "KZN",

		"long_name": "Gauteng",
		"short_name": "GP",

		"long_name": "Free State",
		"short_name": "FS",

		"long_name": "Limpopo",
		"short_name": "LP",

		"long_name": "Mpumalanga",
		"short_name": "MP",

		"long_name": "North West",
		"short_name": "NW",

		"long_name": "Western Cape",
		"short_name": "WC",

		"long_name": "Eastern Cape",
		"short_name": "EC",

		"long_name": "Northern Cape",
		"short_name": "NC",
	*/

	newCountry = oldCountry
	if oldCountry == nil || *oldCountry == "South Africa" || len(*oldCountry) == 0 {
		defaultCountry := "ZA"
		newCountry = &defaultCountry
	}

	if *newCountry == "ZA" && oldZone != nil {
		zone := *oldZone
		// Gauteng - GT from uAfrica should be GP for Google
		if zone == "GT" {
			zone = "GP"
		} else if zone == "NT" {
			zone = "KZN"
		}

		zone = string_utils.ReplaceCaseInsensitive(zone, "KwaZulu-Natal", "KZN")
		zone = string_utils.ReplaceCaseInsensitive(zone, "KwaZulu Natal", "KZN")
		zone = string_utils.ReplaceCaseInsensitive(zone, "Gauteng", "GP")
		zone = string_utils.ReplaceCaseInsensitive(zone, "Freestate", "FS")
		zone = string_utils.ReplaceCaseInsensitive(zone, "Free State", "FS")
		zone = string_utils.ReplaceCaseInsensitive(zone, "Limpopo", "LP")
		zone = string_utils.ReplaceCaseInsensitive(zone, "Mpumalanga", "MP")
		zone = string_utils.ReplaceCaseInsensitive(zone, "North West", "NW")
		zone = string_utils.ReplaceCaseInsensitive(zone, "Eastern Cape", "EC")
		zone = string_utils.ReplaceCaseInsensitive(zone, "Western Cape", "WC")
		zone = string_utils.ReplaceCaseInsensitive(zone, "Northern Cape", "NC")
		zone = string_utils.ReplaceCaseInsensitive(zone, "Eastern-Cape", "EC")
		zone = string_utils.ReplaceCaseInsensitive(zone, "Western-Cape", "WC")
		zone = string_utils.ReplaceCaseInsensitive(zone, "Northern-Cape", "NC")

		newZone = &zone
	}

	return
}
