package handler_utils

import (
	"reflect"

	"gitlab.com/uafrica/go-utils/errors"
)

type Handler struct {
	RequestParamsType reflect.Type
	RequestBodyType   reflect.Type
	ResponseType      reflect.Type
	FuncValue         reflect.Value
}

type SQSHandler struct {
	RecordType reflect.Type
	FuncValue  reflect.Value
}

func NewHandler(handlerFunction interface{}) (Handler, error) {
	h := Handler{}

	handlerFunctionType := reflect.TypeOf(handlerFunction)
	if handlerFunctionType.NumIn() < 1 || handlerFunctionType.NumIn() > 2 {
		return h, errors.Errorf("takes %d args instead of (Params[, Body])", handlerFunctionType.NumIn())
	}
	if handlerFunctionType.NumOut() < 1 || handlerFunctionType.NumOut() > 2 {
		return h, errors.Errorf("returns %d results instead of ([Response,] error)", handlerFunctionType.NumOut())
	}

	//arg[0] must be a struct for params. It may be an empty struct, but
	//all public fields require a json tag which we will use to math the URL param name
	if err := validateStructType(handlerFunctionType.In(0)); err != nil {
		return h, errors.Wrapf(err, "first arg %v is not valid params struct type", handlerFunctionType.In(0))
	}
	h.RequestParamsType = handlerFunctionType.In(0)

	//arg[1] is optional and must be a struct for request body. It may be an empty struct, but
	//all public fields require a json tag which we will use to unmarshal the request body from JSON
	if handlerFunctionType.NumIn() >= 2 {
		if handlerFunctionType.In(1).Kind() == reflect.Slice {
			if err := validateStructType(handlerFunctionType.In(1).Elem()); err != nil {
				return h, errors.Errorf("second arg %v is not valid body []struct type", handlerFunctionType.In(1))
			}
		} else {
			if err := validateStructType(handlerFunctionType.In(1)); err != nil {
				return h, errors.Errorf("second arg %v is not valid body struct type", handlerFunctionType.In(1))
			}
		}

		//todo: check special fields for claims, and see if also applies to params struct...
		//AccountID must be int64 or *int64 with tag =???
		//UserID must be int64 or *int64 with tag =???
		//Username must be string with tag =???

		h.RequestBodyType = handlerFunctionType.In(1)
	}

	//last result must be error
	if _, ok := reflect.New(handlerFunctionType.Out(handlerFunctionType.NumOut() - 1)).Interface().(*error); !ok {
		return h, errors.Errorf("last result %v is not error type", handlerFunctionType.Out(handlerFunctionType.NumOut()-1))
	}

	h.FuncValue = reflect.ValueOf(handlerFunction)
	return h, nil
}

func validateStructType(t reflect.Type) error {
	if t.Kind() != reflect.Struct {
		return errors.Errorf("%v is %v, not a struct", t, t.Kind())
	}
	for i := 0; i < t.NumField(); i++ {
		f := t.Field(i)
		if f.Name[0] >= 'a' && f.Name[0] <= 'z' {
			//lowercase fields should not have json tag
			if f.Tag.Get("json") != "" {
				return errors.Errorf("%s.%s must be uppercase because it has a json tag \"%s\"",
					t.Name(),
					f.Name,
					f.Tag.Get("json"))
			}
		}

		// 	if f.... check tags recursively... for now, not too strict ... add checks if we see issues that break the API, to help dev to fix before we deploy, or to prevent bad habits...
	}
	return nil
}

func NewSQSHandler(fnc interface{}) (SQSHandler, error) {
	h := SQSHandler{}

	fncType := reflect.TypeOf(fnc)
	if fncType.NumIn() != 1 {
		return h, errors.Errorf("takes %d args instead of (Record)", fncType.NumIn())
	}
	if fncType.NumOut() != 1 {
		return h, errors.Errorf("returns %d results instead of (error)", fncType.NumOut())
	}

	// Arg[0] must be a struct for the message record body.
	if fncType.In(0).Kind() != reflect.Struct {
		return h, errors.Errorf("first arg %v is not valid record struct type", fncType.In(0))
	}
	h.RecordType = fncType.In(0)

	// Result must be error
	if _, ok := reflect.New(fncType.Out(0)).Interface().(*error); !ok {
		return h, errors.Errorf("result %v is not error type", fncType.Out(0))
	}

	h.FuncValue = reflect.ValueOf(fnc)
	return h, nil
}
