package reflection_test

import (
	"encoding/json"
	"reflect"
	"testing"

	"gitlab.com/uafrica/go-utils/reflection"
)

func Test1(t *testing.T) {
	doc := `
		{
				"took":872,
				"timed_out":false,
				"_shards":{
					"total":38,
					"successful":38,
					"skipped":0,
					"failed":0
				},
				"hits":{
					"total":{
						"value":10,
						"relation":"eq"
					},
					"max_score":null,
					"hits":[
						{
							"_index": "go-utils-audit-test-20211030",
							"_type": "_doc",
							"_id": "Tj9l5XwBWRiAneoYazic",
							"_score": 1.2039728,
							"_source": {
								"@timestamp": "2021-10-30T15:03:20.000000+02:00",
								"@end_time": "2021-10-30T15:03:21.000000+02:00",
								"@duration_ms": 1000,
								"test1": "6",
								"test2": "ACC_00098",
								"test3": 10,
								"http": {
									"method": "GET",
									"path": "/accounts"
								},
								"http_method": "GET",
								"http_path": "/accounts"
							}
						}
					]
				}
			}
			`

	//using default type, documents in _source:{} are parsed to map[string]interface{}
	res := SearchResponseBody{}
	if err := json.Unmarshal([]byte(doc), &res); err != nil {
		t.Fatalf("cannot unmarshal response into default type: %+v", err)
	}
	for hitIndex, hit := range res.Hits.Hits {
		t.Logf("doc[%d]: (%T)%+v", hitIndex, hit.Source, hit.Source)
	}

	//create type with own document type to use for _source field:
	cloned, err := reflection.CloneType(
		reflect.TypeOf(SearchResponseBody{}),
		map[string]reflect.Type{
			".hits.hits[]._source": reflect.TypeOf(myDoc{}),
		})
	if err != nil {
		t.Fatal(err)
	}
	t.Logf("cloned type: %v", cloned)

	//unmarshal using new type to have correct type for each hit
	resPtrValue := reflect.New(cloned)
	if err := json.Unmarshal([]byte(doc), resPtrValue.Interface()); err != nil {
		t.Fatalf("failed to decode into cloned type: %+v", err)
	}
	// clonedRes := resPtrValue.Interface()
	// t.Logf("Coned res: %+v", clonedRes)

	//get the replaced values as an array of docs
	v, err := reflection.Get(resPtrValue, ".hits.hits[]._source")
	if err != nil {
		t.Fatalf("Did not get list: %+v", err)
	}
	docs, ok := v.Interface().([]myDoc)
	if !ok {
		t.Fatalf("%T is not []myDoc", v.Interface())
	}
	if len(docs) != 1 {
		t.Fatalf("Got %d != 1", len(docs))
	}
	for _, doc := range docs {
		t.Logf("doc: (%T)%+v", doc, doc)
	}
}

type SearchResponseBody struct {
	Took     int                  `json:"took"` //milliseconds
	TimedOut bool                 `json:"timed_out"`
	Shards   SearchResponseShards `json:"_shards"`
	Hits     SearchResponseHits   `json:"hits"`
}

type SearchResponseShards struct {
	Total      int `json:"total"`
	Successful int `json:"successful"`
	Skipped    int `json:"skipped"`
	Failed     int `json:"failed"`
}

type SearchResponseHits struct {
	Total    SearchResponseHitsTotal `json:"total"`
	MaxScore *float64                `json:"max_score,omitempty"`
	Hits     []HitDoc                `json:"hits"`
}

type SearchResponseHitsTotal struct {
	Value    int    `json:"value"`    //e.g. 0 when no docs matched
	Relation string `json:"relation"` //e.g. "eq"
}

type HitDoc struct {
	Index  string                 `json:"_index"` //name of index
	Type   string                 `json:"_type"`  //_doc
	ID     string                 `json:"_id"`
	Score  float64                `json:"_score"`  //
	Source map[string]interface{} `json:"_source"` //the document of itemType
}

type myDoc struct {
	Test1 string `json:"test1"`
}
