package search_test

import (
	"fmt"
	"math/rand"
	"strings"
	"testing"
	"time"

	"github.com/google/uuid"
	"gitlab.com/uafrica/go-utils/logs"
	"gitlab.com/uafrica/go-utils/search"
)

func TestLocalDocuments(t *testing.T) {
	testDocuments(t, search.Config{
		Addresses: []string{"https://localhost:9200"},
	})
}

func TestDevDocuments(t *testing.T) {
	testDocuments(t, search.Config{
		Addresses: []string{"https://search-uafrica-v3-api-api_logs-fefgiypvmb3sg5wqohgsbqnzvq.af-south-1.es.amazonaws.com/"}, //from AWS Console OpenSearch Service > Domains > uafrica-v3-api-api_logs > General Information: Domain Endpoints
		Username:  "uafrica",
		Password:  "Aiz}a4ee",
	})
}

func testDocuments(t *testing.T, c search.Config) {
	logs.SetGlobalFormat(logs.NewConsole())
	logs.SetGlobalLevel(logs.LevelDebug)
	a, err := search.New(c)
	if err != nil {
		t.Fatalf("failed to create writer: %+v", err)
	}

	indexName := "go-utils-search-docs-test"
	ds, err := a.DocumentStore(indexName, SearchOrder{})
	if err != nil {
		t.Fatalf("failed to create document store: %+v", err)
	}

	//write N documents
	buyers := []string{"Joe", "Anne", "Griet", "Kobus", "Caleb", "Roger", "Susan", "Maria", "Sandra"}
	sellers := []string{"Hannelie", "Angelica", "Louis", "Bertus", "Bongi", "Vusi", "Andrew", "Joseph"}
	nextItem := 0
	itemInfos := []testItemInfo{
		{Name: "Samsung 17\" LCD Monitor", Cost: 0},
		{Name: "Acer 15\" LED Monitor", Cost: 0},
		{Name: "Apple M1 16\" MAC", Cost: 0},
		{Name: "Red Dress size M", Cost: 0},
		{Name: "Grey Shorts size 115cm", Cost: 0},
		{Name: "Black Student Prince School Shoes Boys Size 8", Cost: 0},
		{Name: "Black Student Prince School Shoes Boys Size 9", Cost: 0},
		{Name: "Black Student Prince School Shoes Boys Size 10", Cost: 0},
		{Name: "Black Student Prince School Shoes Boys Size 11", Cost: 0},
		{Name: "Black Student Prince School Shoes Girst Size 6", Cost: 0},
		{Name: "Black Student Prince School Shoes Girst Size 7", Cost: 0},
		{Name: "Black Student Prince School Shoes Girst Size 8", Cost: 0},
		{Name: "Faber Castell HB Pencil", Cost: 0},
		{Name: "Faber Castell 2H Pencil", Cost: 0},
		{Name: "Faber Castell 4H Pencil", Cost: 0},
		{Name: "12 Colour Crayons", Cost: 0},
		{Name: "Steadler Rubber", Cost: 0},
	}
	N := 100
	for i := 0; i < N; i++ {
		//make a random document
		id := uuid.New().String()
		buyer := buyers[rand.Intn(len(buyers))]
		seller := sellers[rand.Intn(len(sellers))]
		doc := SearchOrder{
			ID:                            int64(i + 1),
			AccountID:                     int64(i + 2),
			AccountOrderNumber:            int64(i * 2),
			ChannelID:                     int64(i),
			ChannelOrderNumber:            fmt.Sprintf("CHO-%d", i),
			ChannelOrderReference:         fmt.Sprintf("REF-%05d", i),
			CustomerName:                  buyer,
			CustomerEmail:                 strings.ToLower(buyer + "@home.net"),
			CustomerPhone:                 "123456789",
			DeliveryAddress:               "My Street",
			Currency:                      "ZAR",
			Items:                         []SearchOrderItem{},
			TotalPrice:                    0,
			TotalWeightKg:                 0,
			TotalQty:                      0,
			TotalFulfilledQty:             0,
			Status:                        OrderStatusNew,
			PaymentStatus:                 OrderPaymentStatusUnpaid,
			TimeCreated:                   time.Now(),
			TimeModified:                  nil,
			Tags:                          "1(blue),2(red)",
			BuyerSelectedShippingCost:     1.23,
			BuyerSelectedShippingProvider: seller,
		}
		for i := 0; i < rand.Intn(5)+1; i++ {
			itemInfo := itemInfos[nextItem]
			nextItem++
			if nextItem >= len(itemInfos) {
				nextItem = 0
			}
			item := SearchOrderItem{
				SKU:         fmt.Sprintf("SKU-%s-%03d", itemInfo.Name[:3], i),
				Description: itemInfo.Name,
				UnitPrice:   itemInfo.Cost,
				Qty:         rand.Intn(3) + 1,
			}
			doc.Items = append(doc.Items, item)
			doc.TotalPrice += item.UnitPrice * float64(item.Qty)
			doc.TotalQty += item.Qty
		}

		//add to the search
		if err := ds.Write(id, doc); err != nil {
			t.Fatalf("failed to add doc: %+v", err)
		}
	}

	//search some of the documents
	query := search.Query{
		MultiMatch: &search.QueryMultiMatch{
			Query:  buyers[0],
			Fields: []string{"buyer"},
		},
	}
	ids, totalCount, err := ds.Search(query, 10)
	if err != nil {
		t.Errorf("failed to search: %+v", err)
	} else {
		t.Logf("search result total_count:%d with %d ids", totalCount, len(ids))
		if len(ids) > 10 {
			t.Errorf("got %d docs > max 10", len(ids))
		}
		for _, id := range ids {
			if getDoc, err := ds.Get(id); err != nil {
				t.Fatalf("failed to get: %+v", err)
			} else {
				t.Logf("   GOT: %s: %+v", id, getDoc)
			}
		}
	}
	t.Logf("Done")
}

// type testDocument struct {
// 	//UUID       string    `json:"uuid"`
// 	Buyer      string    `json:"buyer" search:"keyword"`
// 	Seller     string    `json:"seller" search:"keyword"`
// 	Items      []docItem `json:"items"`
// 	TotalCost  float64   `json:"total_cost"`
// 	TotalItems int       `json:"total_items"`
// }

// type docItem struct {
// 	Description string  `json:"description" search:"keyword"`
// 	UnitCost    float64 `json:"unit_cost"`
// 	Qty         int     `json:"qty"`
// }

type testItemInfo struct {
	Name string
	Cost float64
}

type SearchOrder struct {
	ID                            int64              `json:"id"`
	AccountID                     int64              `json:"account_id"                search:"keyword"`
	AccountOrderNumber            int64              `json:"account_order_number"`
	ChannelID                     int64              `json:"channel,omitempty"         search:"keyword"`
	ChannelOrderNumber            string             `json:"channel_order_number,omitempty"`
	ChannelOrderReference         string             `json:"channel_order_reference,omitempty"`
	CustomerName                  string             `json:"customer_name,omitempty"`
	CustomerEmail                 string             `json:"customer_email,omitempty"`
	CustomerPhone                 string             `json:"customer_phone,omitempty"`
	DeliveryAddress               string             `json:"delivery_address,omitempty"`
	Currency                      string             `json:"currency"                  search:"keyword"`
	Items                         []SearchOrderItem  `json:"items,omitempty"`
	TotalPrice                    float64            `json:"total_price"`
	TotalWeightKg                 float64            `json:"total_weight_kg"`
	TotalQty                      int                `json:"total_qty"`
	TotalFulfilledQty             int                `json:"total_fulfilled_qty"`
	Status                        OrderStatus        `json:"status"                    search:"keyword"`
	PaymentStatus                 OrderPaymentStatus `json:"payment_status"            search:"keyword"`
	TimeCreated                   time.Time          `json:"time_created"`
	TimeModified                  *time.Time         `json:"time_modified,omitempty"`
	Tags                          string             `json:"tags"                      search:"keyword"` //CSV or tags sorted, so [A,B] and [B,A] both -> "A,B" keyword
	BuyerSelectedShippingCost     float64            `json:"buyer_selected_shipping_cost,omitempty"`
	BuyerSelectedShippingProvider string             `json:"buyer_selected_shipping_provider,omitempty"`
}

type SearchOrderItem struct {
	SKU           string  `json:"sku"                          search:"keyword"`
	Description   string  `json:"description"                  search:"keyword"`
	Vendor        string  `json:"vendor"                       search:"keyword"`
	UnitPrice     float64 `json:"unit_price"`
	UnitWeightKg  float64 `json:"unit_weight_kg"`
	Qty           int     `json:"qty"`
	FulfilledQty  int     `json:"fulfilled_qty"`
	TotalPrice    float64 `json:"total_price"`
	TotalWeightKg float64 `json:"total_weight_kg"`
}

type OrderStatus string

const (
	OrderStatusNew       OrderStatus = "new"
	OrderStatusCompleted OrderStatus = "completed"
	OrderStatusCancelled OrderStatus = "cancelled"
)

type OrderPaymentStatus string

const (
	OrderPaymentStatusUnpaid            OrderPaymentStatus = "unpaid"
	OrderPaymentStatusPending           OrderPaymentStatus = "pending"
	OrderPaymentStatusPartiallyPaid     OrderPaymentStatus = "partially-paid"
	OrderPaymentStatusPaid              OrderPaymentStatus = "paid"
	OrderPaymentStatusPartiallyRefunded OrderPaymentStatus = "partially-refunded"
	OrderPaymentStatusRefunded          OrderPaymentStatus = "refunded"
	OrderPaymentStatusVoided            OrderPaymentStatus = "voided"
	OrderPaymentStatusAuthorised        OrderPaymentStatus = "authorised"
)
