package struct_utils_test

import (
	"encoding/json"
	"fmt"
	"net/url"
	"os"
	"strings"
	"testing"
	"time"

	"gitlab.com/uafrica/go-utils/errors"
	"gitlab.com/uafrica/go-utils/logs"
	"gitlab.com/uafrica/go-utils/struct_utils"
)

func TestEnv(t *testing.T) {
	logs.SetGlobalFormat(logs.NewConsole())
	logs.SetGlobalLevel(logs.LevelDebug)
	//booleans
	os.Setenv("TEST_VALUE_ENABLE_CACHE", "true")
	os.Setenv("TEST_VALUE_DISABLE_LOG", "true")
	os.Setenv("TEST_VALUE_ADMIN", "false")

	//integers
	os.Setenv("TEST_VALUE_MAX_SIZE", "12")

	os.Setenv("TEST_VALUE_SEQ1", "[4,5,6]") //list in one value

	os.Setenv("TEST_VALUE_SEQ2_10", "10") //numbered list elements
	os.Setenv("TEST_VALUE_SEQ2_20", "20")
	os.Setenv("TEST_VALUE_SEQ2_4", "4")
	os.Setenv("TEST_VALUE_SEQ2_15", "15")
	os.Setenv("TEST_VALUE_SEQ2", "100")

	os.Setenv("TEST_VALUE_CUTOFF", "2021-11-20T12:00:00+02:00")
	os.Setenv("TEST_VALUE_HOLIDAYS", "[2021-03-21,2021-04-27,2021-05-01,2021-06-16,2021-08-09,2021-12-16,2021-12-25]")

	//=====[ TEST THIS FUNCTION ]=====
	nv := struct_utils.NamedValuesFromEnv("TEST_VALUE")

	testNamedValues(t, nv)
}

func TestURL1(t *testing.T) {
	logs.SetGlobalFormat(logs.NewConsole())
	logs.SetGlobalLevel(logs.LevelDebug)

	queryParams := map[string]string{
		"enable_cache": "true",
		"disable_log":  "true",
		"admin":        "false",
		"max_size":     "12",
		"seq1":         "[4,5,6]",
		"seq2":         "[100,4,10,15,20]", //url does not support _# numbering of params in a list, only [csv] or multi-value with seq2=100&seq2=4&... testing in TestURL2()
		"cutoff":       "2021-11-20T12:00:00+02:00",
		"holidays":     "[2021-03-21,2021-04-27,2021-05-01,2021-06-16,2021-08-09,2021-12-16,2021-12-25]",
	}

	//=====[ TEST THIS FUNCTION ]=====
	nv := struct_utils.NamedValuesFromURL(queryParams, nil)
	testNamedValues(t, nv)
}

func TestURL2(t *testing.T) {
	logs.SetGlobalFormat(logs.NewConsole())
	logs.SetGlobalLevel(logs.LevelDebug)

	queryParams := map[string]string{
		"disable_log": "true",
		"admin":       "false",
		"max_size":    "12",
		"seq1":        "4",
		"cutoff":      "2021-11-20T12:00:00+02:00",
	}
	multiValueParams := map[string][]string{
		"enable_cache": {"true"},
		"seq1":         {"5", "6"}, //merged with above "4"
		"seq2":         {"100", "4", "10", "15", "20"},
		"holidays":     {"2021-03-21", "2021-04-27", "2021-05-01", "2021-06-16", "2021-08-09", "2021-12-16", "2021-12-25"},
	}

	//=====[ TEST THIS FUNCTION ]=====
	nv := struct_utils.NamedValuesFromURL(queryParams, multiValueParams)
	testNamedValues(t, nv)
}

func TestURL3(t *testing.T) {
	urlString := "/test?admin=false&cutoff=2021-11-20T12%3A00%3A00%2B02%3A00&disable_log=true&enable_cache=true&holidays=2021-03-21&holidays=2021-04-27&holidays=2021-05-01&holidays=2021-06-16&holidays=2021-08-09&holidays=2021-12-16&holidays=2021-12-25&max_size=12&seq1=4&seq1=5&seq1=6&seq2=100&seq2=4&seq2=10&seq2=15&seq2=20"
	u, err := url.Parse(urlString)
	if err != nil {
		t.Fatalf("cannot parse URL")
	}
	nv := struct_utils.NamedValuesFromURL(nil, u.Query())
	testNamedValues(t, nv)
}

func testNamedValues(t *testing.T, nv map[string][]string) {
	//assets expected values
	exp := map[string][]string{
		"enable_cache": {"true"},
		"disable_log":  {"true"},
		"admin":        {"false"},
		"max_size":     {"12"},
		"seq1":         {"4", "5", "6"},
		"seq2":         {"100", "4", "10", "15", "20"}, //order is important
		"cutoff":       {"2021-11-20T12:00:00+02:00"},
		"holidays":     {"2021-03-21", "2021-04-27", "2021-05-01", "2021-06-16", "2021-08-09", "2021-12-16", "2021-12-25"},
	}
	if len(nv) != len(exp) {
		t.Fatalf("len(nv)=%d != len(exp)=%d: %+v", len(nv), len(exp), nv)
	}

	for name, expValues := range exp {
		if envValues, ok := nv[name]; !ok {
			t.Fatalf("%s not set in %+v", name, nv)
		} else {
			if len(expValues) != len(envValues) {
				t.Fatalf("%s has %d != %d values (%v != %v)", name, len(envValues), len(expValues), envValues, expValues)
			}
			for i, v := range envValues {
				if v != expValues[i] {
					t.Fatalf("%s[%d] = \"%s\" != \"%s\"", name, i, v, expValues[i])
				}
			}
		}
	}

	//=====[ PARSE INTO STRUCT ]==========
	c := Config{}
	unused, err := struct_utils.UnmarshalNamedValues(nv, &c)
	if err != nil {
		t.Fatalf("failed: %+v", err)
	}
	if len(unused) != 0 {
		t.Fatalf("unused: %+v", unused)
	}
	t.Logf("parsed struct: %+v", c)

	{
		e := json.NewEncoder(os.Stdout)
		e.SetIndent("", "\t")
		e.Encode(c)
	}

	if !c.EnableCache || !c.DisableLog || c.Admin {
		t.Fatalf("wrong bool values: %+v", c)
	}
	if c.MaxSize != 12 {
		t.Fatalf("wrong nr values: %+v", c)
	}
	if len(c.Seq1) != 3 || c.Seq1[0] != 4 || c.Seq1[1] != 5 || c.Seq1[2] != 6 {
		t.Fatalf("wrong seq1: %+v", c)
	}
	if len(c.Seq2) != 5 || c.Seq2[0] != 100 || c.Seq2[1] != 4 || c.Seq2[2] != 10 || c.Seq2[3] != 15 || c.Seq2[4] != 20 {
		t.Fatalf("wrong seq2: %+v", c)
	}
	if c.Cutoff.UTC().Format("2006-01-02 15:04:05") != "2021-11-20 10:00:00" {
		t.Fatalf("wrong cutoff")
	}
	if len(c.Holidays) != 7 ||
		c.Holidays[0].String() != "2021-03-21" ||
		c.Holidays[1].String() != "2021-04-27" ||
		c.Holidays[2].String() != "2021-05-01" ||
		c.Holidays[3].String() != "2021-06-16" ||
		c.Holidays[4].String() != "2021-08-09" ||
		c.Holidays[5].String() != "2021-12-16" ||
		c.Holidays[6].String() != "2021-12-25" {
		t.Fatalf("wrong holidays")
	}
}

type Config struct {
	EnableCache bool      `json:"enable_cache"`
	DisableLog  bool      `json:"disable_log"`
	Admin       bool      `json:"admin"`
	MaxSize     int64     `json:"max_size"`
	Seq1        []int     `json:"seq1"`
	Seq2        []int64   `json:"seq2"`
	Cutoff      time.Time `json:"cutoff"`
	Holidays    []Date    `json:"holidays"`
}

type Date struct {
	Y, M, D int
}

func (d *Date) Scan(value []byte) error {
	s := strings.Trim(string(value), "\"")
	v, err := time.ParseInLocation("2006-01-02", s, time.Now().Location())
	if err != nil {
		return errors.Errorf("%s is not CCYY-MM-DD", s)
	}
	d.Y = v.Year()
	d.M = int(v.Month())
	d.D = v.Day()
	return nil
}

func (d *Date) UnmarshalJSON(value []byte) error {
	return d.Scan(value)
}

func (d Date) String() string {
	return fmt.Sprintf("%04d-%02d-%02d", d.Y, d.M, d.D)
}

func (d Date) MarshalJSON() ([]byte, error) {
	return []byte("\"" + d.String() + "\""), nil
}
